"""LookML Generator implementations."""

import os
from typing import Dict
from looker_loader.models.looker import LookerView, ValidatedLookerDimension, LookerMeasure
import logging


class LookmlGenerator:
    """
        Main LookML generator that coordinates dimension, view, and explore generation.
    """
    def __init__(self, cli_args):
        self._cli_args = cli_args
    
    def _generate_views(self, model, views = None, parent = None):
        """Split up the model into views"""

        if views is None:
            views = []

        view_dimensions = []
        view_measures = []

        if parent is not None:
            view_name = f"{parent}__{model.name}"
        else:
            view_name = model.name

        for field in model.fields:
            view_dimensions.append(field.model_dump())

            if field.fields is not None:
                self._generate_views(field, views, parent=view_name)
            if field.measures is not None:
                for measure in field.measures:
                    view_measures.append(measure.model_dump())
        view = LookerView(**{
            "name": view_name,
            "sql_table_name": model.sql_table_name,
            "dimensions": view_dimensions,
            "measures": view_measures,
        })
        views.append(view)

        return views

    def _generate_joins(self, model, joins = None, parent = None, depth = 0):
        """Recursively create joins for a model by detecting nested fields."""
        if joins is None:
            joins = []

        if parent is not None:
            join_name = f"{parent}__{model.name}"
        else:
            join_name = model.name

        for field in model.fields:
            if field.fields is not None:
                self._generate_joins(field, joins, parent=join_name, depth=depth + 1)

        if parent is not None:
            join = {
                "name": join_name,
                "sql": f"LEFT JOIN UNNEST(${{{parent}.{model.name}}}) AS {join_name}",
                "type": "left_outer",
                "relationship": "one_to_many",
                "required_joins": parent if depth > 1 else None,
            }
            joins.append(join)

        return joins

    def _create_explore(self, model):
        """Create an explore for the model"""
        joins = self._generate_joins(model)

        if joins:
            explore = {
                "name": model.name,
                "joins": joins,
            }
            return explore

    def generate(self, model) -> Dict:
        """Generate LookML for a model."""
        
        view_groups = self._generate_views(model)

        explore = self._create_explore(model)

        return view_groups, explore


from tests.fixtures.bigquery import bigquery_schema_1

def test_bigquery_schema(bigquery_schema_1):
    """Test the BigQuery schema"""
    assert bigquery_schema_1 is not None
    assert len(bigquery_schema_1.fields) == 2
    assert bigquery_schema_1.fields[0].name == "pk_obt"
    assert bigquery_schema_1.fields[1].name == "weekly_data"
    assert len(bigquery_schema_1.fields[1].fields) == 5
    assert bigquery_schema_1.fields[1].fields[0].name == "question_cohort_week_id"
    assert bigquery_schema_1.fields[1].fields[3].name == "number_of_people"
    assert bigquery_schema_1.fields[1].fields[4].name == "answer_value_response_information"
    assert len(bigquery_schema_1.fields[1].fields[4].fields) == 2
    assert bigquery_schema_1.fields[1].fields[4].fields[0].name == "question_cohort_week_answer_id"
    assert bigquery_schema_1.fields[1].fields[4].fields[1].name == "answer_value"



from pydantic import BaseModel, model_validator
from typing import List, Optional, Dict
from looker_loader.enums import (
    LookerType,
    LookerBigQueryDataType
)
# Models for database information loading

class DatabaseField(BaseModel):
    name: str
    type: LookerType
    db_type: str
    order: int
    mode: Optional[str] = None
    description: Optional[str] = None
    parent_name: Optional[str] = None
    parent_mode: Optional[str] = None
    parent_type: Optional[str] = None
    parent_db_type: Optional[str] = None
    sql: Optional[str] = None
    fields: Optional[List["DatabaseField"]] = None

    @model_validator(mode="before")
    def adjust_type(cls, values):
        """Adjust the type of the field based on the db_type."""
        db_type = values.get("type")
        type = LookerBigQueryDataType.get(db_type.upper())
        if type is None:
            raise ValueError(f"Invalid type: {db_type}")
        values["type"] = type
        values["db_type"] = db_type.upper()
        return values

    @model_validator(mode="before")
    def push_down_copy_for_repeated(cls, values):
        """ push down copy for repeated fields that are arrays """
        if values.get("mode") == "REPEATED" and values.get("fields") is None:
            copy = values.copy()
            copy["mode"] = "NULLABLE"
            copy["parent_db_type"] = "ARRAY"
            copy["parent_mode"] = "REPEATED"
            copy["description"] = f"A single value from {values.get("description", "")}"
            values["fields"] = [copy]
        return values

    @model_validator(mode="before")
    def push_down_attributes(cls, values):
        """ push parent_name, parent_mode and parent_type to children """
        if values.get("fields") is not None:
            inherited_children = []
            for child in values.get("fields"):
                child["parent_mode"] = values.get("mode")
                child["parent_type"] = values.get("type")
                child["parent_db_type"] = values.get("db_type")
                if hasattr(values, "parent_name"):
                    child["parent_name"] = f"{values.get('parent_name')}.{values.get('name')}"
                else:
                    child["parent_name"] = values.get("name")
                inherited_children.append(DatabaseField(**child))
            values["fields"] = inherited_children
        return values
    
    @model_validator(mode="before")
    def push_order(cls, values):
        """ push order to children 
            order of the fields from the parent table
            is used to order the children fields
        """
        if values.get("fields") is not None:
            fields = []
            for i, field in enumerate(values.get("fields")):
                field["order"] = i
                fields.append(field)
            values["fields"] = fields
        return values

    @model_validator(mode="after")
    def create_sql(cls, values):
        """Create SQL field from name and parent_name"""
        base = "${TABLE}"
        if values.parent_mode == "REPEATED" and values.parent_db_type == "ARRAY":
            values.sql = base
        else:
            values.sql = f"{base}.{values.name}"
        return values

    class Config:
        from_attributes = True


class DatabaseTable(BaseModel):
    name: str
    fields: List[DatabaseField]
    type: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    sql_table_name: Optional[str] = None

    table_group: Optional[str] = None
    table_project: Optional[str] = None

    class Config:
        from_attributes = True

    @model_validator(mode="before")
    def push_order(cls, values):
        """ push order to children 
            order of the fields from the parent table
            is used to order the children fields
        """
        if values.get("fields") is not None:
            fields = []
            for i, field in enumerate(values.get("fields")):
                field["order"] = i
                fields.append(field)
            values["fields"] = fields
        return values




import lkml
import logging
import re

def fix_multiline_indentation(text, indent_size=2):
    """Fixes indentation for multiline strings in LookML."""
    lines = text.split('\n')
    fixed_lines = []
    inside_multiline_string = False
    current_indent = ""

    for line in lines:
        stripped = line.lstrip()
        leading_spaces = ' ' * (len(line) - len(stripped))

        # If not currently inside a multiline string
        if not inside_multiline_string:
            fixed_lines.append(line)
            # Check if a string starts but does not end on the same line
            quote_starts = stripped.count('"') % 2 == 1
            if quote_starts:
                inside_multiline_string = True
                current_indent = leading_spaces + (' ' * indent_size)
        else:
            # Inside a multiline string
            fixed_lines.append(current_indent + stripped)
            # Check if string ends on this line
            if stripped.count('"') % 2 == 1:
                inside_multiline_string = False

    return '\n'.join(fixed_lines)

def remove_empty_from_dict(data):
  """Recursively removes keys with None or empty list values from a dictionary.

  Args:
    data: The dictionary to clean.

  Returns:
    A new dictionary with empty values removed.
  """
  if isinstance(data, dict):
    return {
        k: remove_empty_from_dict(v)
        for k, v in data.items()
        if v is not None and (not isinstance(v, list) or len(v) > 0)
    }
  elif isinstance(data, list):
    return [remove_empty_from_dict(item) for item in data if item is not None and (not isinstance(item, list) or len(item) > 0)]
  else:
    return data

def convert_to_lkml(views, explore):
    """Convert a dictionary to a LookML string."""
        
    python_r = []
    for view in views:
        python_r.append(remove_empty_from_dict(view.dict(exclude_none=True)))

    if explore is not None:
      dumpfile = {
        "explore": remove_empty_from_dict(explore),
        "views": python_r,
      }
    else:
      dumpfile = {
          "views": python_r,
      }

    lk = lkml.dump(dumpfile)
    indention_fixed = fix_multiline_indentation(lk)

    return indention_fixed


from looker_loader.enums import ExtendedEnum

class BigqueryUrl(ExtendedEnum):
    BIGQUERY = "https://bigquery.googleapis.com/bigquery/v2/projects/{project_id}/datasets/{dataset_id}/tables/{table_id}"

class BigqueryMode(ExtendedEnum):
    REPEATED = "REPEATED"

class BigqueryType(ExtendedEnum):
    RECORD = "RECORD"
    ARRAY = "ARRAY"
    STRUCT = "STRUCT"


from looker_loader.models.recipe import LookerMixture, LookerMixtureDimension
from typing import List, Set
import json
import Levenshtein
def gather_field_names(mixtures: List[LookerMixture]) -> Set[str]:
    """
    Gather field names from Looker mixtures and their dimensions.
    """
    field_names = []
    
    for mixture in mixtures:
        for dimension in mixture.fields:
            field_names.append(dimension.name)

            if dimension.measures is not None:
                for measure in dimension.measures:
                    field_names.append(measure.name)
        

    return set(field_names)

def group_strings_by_similarity(string_list, group_size=10, similarity_threshold=0.8):
    """
    Groups strings into clusters of similar strings based on Levenshtein distance.

    Args:
        string_list: A list of strings.
        group_size: The desired size of each group.
        similarity_threshold: A threshold for similarity (0.0 to 1.0). Higher = more similar.
                              This is a percentage, calculated as (max_distance - distance) / max_distance

    Returns:
        A list of lists, where each inner list is a group of similar strings.
    """

    groups = []
    unassigned = list(string_list)  # Copy of the list to track unassigned strings

    while unassigned:
        # Start a new group with the first unassigned string
        group = [unassigned.pop(0)]  # Remove and take the first element

        # Find the most similar strings to the first string in the group and add them
        while len(group) < group_size and unassigned:
            best_match = None
            best_similarity = 0

            for i, s in enumerate(unassigned):
                distance = Levenshtein.distance(group[0], s)
                max_distance = max(len(group[0]), len(s))  # Normalize by length
                similarity = (max_distance - distance) / max_distance  # Calculate similarity as a percentage

                if similarity > best_similarity:
                    best_similarity = similarity
                    best_match = i

            if best_match is not None and best_similarity >= similarity_threshold:
                group.append(unassigned.pop(best_match))
            else:
                break  # No more good matches found

        groups.append(group)

    return groups


import argparse
import os
import logging
import lkml
from rich.logging import RichHandler
from looker_loader.utils import FileHandler
from looker_loader.models.recipe import CookBook
from looker_loader.models.config import Config
from looker_loader.databases.bigquery.database import BigQueryDatabase
from looker_loader.tools import recipe_mixer
from looker_loader.models.recipe import LookerMixture
from looker_loader.generator.lookml import LookmlGenerator
from looker_loader.tools.lkml_converter import convert_to_lkml
import asyncio

logging.basicConfig(
    level=logging.INFO, format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
)

class Cli:
    HEADER = """
    Load your data into looker
    """

    def __init__(self):
        self._args_parser = self._init_argparser()
        self._file_handler = FileHandler()

    def _init_argparser(self):
        """Create and configure the argument parser"""
        parser = argparse.ArgumentParser(
            description=self.HEADER,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        parser.set_defaults(
            build_explore=True, write_output=True, hide_arrays_and_structs=True
        )
        parser.add_argument(
            "--table","-t",
            help="The name of the table to generate LookML for",
            type=str,
        )
        parser.add_argument(
            "--config",
            help="Path to the config file",
            type=str,
            default=".",
        )
        # parser.add_argument(
        #     "--version",
        #     action="version",
        #     version=f'looker-loader {version("looker-loader")}',
        # )
        # parser.add_argument(
        #     "--output-dir",
        #     help="Path to a directory that will contain the generated lookml files",
        #     default=self.DEFAULT_LOOKML_OUTPUT_DIR,
        #     type=str,
        # )
        return parser

    def _write_lookml_file(
        self,
        output_dir: str,
        file_path: str,
        contents: str,
    ) -> str:
        """Write LookML content to a file."""

        file_name = os.path.basename(file_path)
        file_path = os.path.join(output_dir, file_path.split(file_name)[0])
        os.makedirs(file_path, exist_ok=True)

        file_path = f"{file_path}/{file_name}"

        # Write contents
        self._file_handler.write(file_path, contents)

        return file_path

    def _load_recipe(self, folder: str = None):
        """Load the recipe from a yaml file"""
        args = self._args_parser.parse_args()
        if folder is None:
            folder = args.config
        if not os.path.exists(folder):
            raise FileNotFoundError(f"Folder {folder} does not exist")

        logging.info(f"Loading Recipe from {folder}/loader_recipe.yml")
        data = self._file_handler.read(f"{folder}/loader_recipe.yml", file_type="yaml")
        self.recipe = CookBook(**data)
        self.mixer = recipe_mixer.RecipeMixer(self.recipe)

    def _load_config(self, folder: str = None):
        """Load the config from a yaml file"""
        args = self._args_parser.parse_args()
        if folder is None:
            folder = args.config
        if not os.path.exists(folder):
            raise FileNotFoundError(f"Folder {folder} does not exist")

        logging.info(f"Loading Config from {folder}/loader_config.yml")
        data = self._file_handler.read(f"{folder}/loader_config.yml", file_type="yaml")
        self.config = Config(**data['config'])

    def _load_tables(self):
        """Load the the list of tables to get schemas for from the database"""
        process_list = []

        for d in self.config.bigquery:
            if not d.tables:
                logging.info(f"Finding all tables in dataset {d.dataset_id} of project {d.project_id}")
                tables = self.database.get_tables_in_dataset(d.project_id, d.dataset_id)
            else:
                tables = d.tables

            for table in tables:
                
                process_list.append(
                    {
                        "project_id": d.project_id,
                        "dataset_id": d.dataset_id,
                        "table_id": table,
                    }
                )

        self.tables = process_list

    async def get_schemas(self):
        """
            asyncronously fetch the schemas of the tables
            and parse them into a common database schema
            and store them in self.schemas
        """
        self.database.init()
        tasks = [
            self.database._async_fetch_table_schema(
            project_id=table.get("project_id"),
            dataset_id=table.get("dataset_id"),
            table_id=table.get("table_id"),
            )
            for table in self.tables
            ]
        
        # Run all tasks concurrently and gather the results
        results = await asyncio.gather(*tasks)

        schemas = []
        for json in results:
            schemas.append(self.database._parse_schema(json))

        self.schemas = schemas

    def run(self):
        """Run the CLI"""
        args = self._args_parser.parse_args()
        self.database = BigQueryDatabase()
        self.lookml = LookmlGenerator(cli_args=args)

        self._load_recipe()
        self._load_config()
        self._load_tables()

        asyncio.run(self.get_schemas())

        mixtures = []
        for schema in self.schemas:
            logging.debug(f"Generating LookML for '{schema.name}'")

            mixture = self.mixer.mixturize(schema)
            mixtures.append(mixture)
        
        from looker_loader.tools.label_gatherer import gather_field_names, group_strings_by_similarity
        import json
        from looker_loader.exceptions import CliError
        from looker_loader.tools.llm import get_field_label

        try:
            lexicanum = self._file_handler.read(f"lexicanum.json", file_type="json")
        except CliError:
            logging.warning("lexicanum.json not found")
            lexicanum = {}
        
        field_names = gather_field_names(mixtures)
        for field_name in field_names:
            if field_name not in lexicanum:
                logging.info(f"Adding field '{field_name}' to lexicanum")
                lexicanum[field_name] = None

        from rich import print 
        print(group_strings_by_similarity(field_names, group_size=20, similarity_threshold=0.3))

        for field, label in lexicanum.items():
            if label is None:
                logging.info(f"Generating label for field '{field}'")
                lexicanum[field] = get_field_label(field_name=field)

        self._file_handler.write("lexicanum.json", json.dumps(lexicanum, indent=2))


        for mixture in mixtures:

            views, explore = self.lookml.generate(
                model=mixture,
            )

            self._write_lookml_file(
                output_dir=f'output/{schema.table_group}',
                file_path=f'{schema.name}.view.lkml',
                contents=convert_to_lkml(views, explore),
            )

        logging.info("LookML files generated successfully")

def main():
    cli = Cli()
    cli.run()

if __name__ == "__main__":
    main()


import google.auth
from google.auth.transport.requests import Request
import requests

from looker_loader.models.database import DatabaseTable

from looker_loader.databases.bigquery.enums import BigqueryMode, BigqueryType, BigqueryUrl
import httpx
import logging

class BigQueryDatabase:
    def __init__(self):
        """Initialize the BigQueryDatabase class."""
        self.database_type = "bigquery"

    def init(self):
        """Authenticate the user with Google Cloud using default credentials."""
        credentials, _ = google.auth.default()
        credentials.refresh(Request())
        self.headers = {
            "Authorization": f"Bearer {credentials.token}",
            "Content-Type": "application/json",
        }

    def _fetch_table_schema(
        self, project_id: str, dataset_id: str, table_id: str
    ) -> DatabaseTable:
        """Fetch the schema of a BigQuery table and parse it into a Pydantic model."""
        self.init()
        url = BigqueryUrl.BIGQUERY.value.format(
            project_id=project_id, dataset_id=dataset_id, table_id=table_id
        )

        response = requests.get(url, headers=self.headers, timeout=10)
        response.raise_for_status()

        self.json_schema = response.json()

    async def _async_fetch_table_schema(self, project_id: str, dataset_id: str, table_id: str):
        """Fetch schema data for a table"""
        # logging.info("Fetching schema for table '%s'", table_id)

        url = BigqueryUrl.BIGQUERY.value.format(
            project_id=project_id, dataset_id=dataset_id, table_id=table_id
        )
        async with httpx.AsyncClient() as client:
            data = await client.get(url, headers=self.headers, timeout=10)  # Await the response and get the content
        return data.json()  # Return the JSON content

    def _parse_schema(self, json) -> DatabaseTable:
        """Parse the schema of a BigQuery table into a Pydantic model."""
        table_ref = json.get("tableReference")
        fields = json.get("schema").get("fields")

        return DatabaseTable(
            name=table_ref.get("tableId"),
            table_group=table_ref.get("datasetId"),
            table_project=table_ref.get("projectId"),
            fields=fields,
            sql_table_name=f'{table_ref.get("projectId")}.{table_ref.get("datasetId")}.{table_ref.get("tableId")}',
        )

    def get_table_schema(self, project_id, dataset_id, table_id) -> DatabaseTable:
        """get the schema of a bigquery table and parse it into a common database schema."""
        logging.info("Fetching schema for table '%s'", table_id)
        json_schema = self._fetch_table_schema(project_id, dataset_id, table_id)
        return self._parse_schema(json_schema)

    def get_tables_in_dataset(self, project_id: str, dataset_id: str) -> list[DatabaseTable]:
        """Get all tables in a BigQuery dataset."""

        from google.cloud import bigquery
        client = bigquery.Client()
        dataset_url = f"{project_id}.{dataset_id}"
        tables = client.list_tables(dataset_url)

        table_list = []
        for table in tables:
            table_list.append(table.table_id)
        return table_list # Make an API request.



from typing import List, Optional, Union
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

# Models for the configuration of Looker Loader

class BigQuery(BaseModel):
    """BigQuery model for Looker Loader"""
    project_id: str
    dataset_id: str
    tables: Optional[List[str]] = Field(default=None, description="List of tables to load")

class Config(BaseModel):
    """Config model for Looker Loader"""
    bigquery: List[BigQuery]


from typing import List, Optional, Union, Literal
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
import warnings
from looker_loader.enums import LookerType, LookerDateTimeframes, LookerTimeTimeframes

from looker_loader.enums import (
    LookerType,
    LookerMeasureType,
    LookerTimeFrame,
    LookerValueFormatName,
)

#  metaclass
class LookerViewElement(BaseModel):
    """Looker data for a view element."""
    name: str = Field(default=None)
    type: Optional[Literal[
        "bin",
        "date",
        "datetime",
        "distance",
        "duration",
        "location",
        "number",
        "string",
        "tier",
        "time",
        "yesno",
        "zipcode",
        "timestamp",
        None
    ]] = None
    label: Optional[str] = None
    hidden: Optional[bool] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None

    value_format_name: Optional[str] = Field(default=None)
    group_label: Optional[str] = None

    @field_validator("value_format_name", mode="before")
    def validate_format_name(cls, value):
        if value is not None:
            if isinstance(value, str):
                value = value.strip()
                if not LookerValueFormatName.has_value(value):
                    warnings.warn(
                        f"Invalid value for value_format_name [{value}]. Setting to None. ",
                    )
                    return None
                else:
                    return LookerValueFormatName(value)
        return value

class LookerMeasureFilter(BaseModel):
    filter_dimension: str
    filter_expression: str

class LookerMeasure(LookerViewElement):
    """Looker data for a measure.
    
    :
        looker:
            measures:
    """

    # Required fields
    type: LookerMeasureType

    # Fields specific to certain measure types
    approximate: Optional[bool] = None  # For count_distinct
    approximate_threshold: Optional[int] = None  # For count_distinct
    precision: Optional[int] = None  # For average, sum
    sql_distinct_key: Optional[str] = None  # For count_distinct
    percentile: Optional[int] = None  # For percentile measures
    required_access_grants: Optional[List[str]] = Field(default=None)
    filters: Optional[List[LookerMeasureFilter]] = None
    sql: Optional[str] = None  # SQL expression for the measure

    @model_validator(mode="before")
    def validate_measure_attributes(cls, values):
        """Validate that measure attributes are compatible with the measure type."""
        if type(values) is dict:
            measure_type = values.get("type")

            if any(
                v is not None
                for v in [
                    values.get("approximate"),
                    values.get("approximate_threshold"),
                    values.get("sql_distinct_key"),
                ]
            ) and measure_type not in (
                LookerMeasureType.COUNT_DISTINCT,
                LookerMeasureType.SUM_DISTINCT,
            ):
                raise ValueError(
                    "approximate, approximate_threshold, and sql_distinct_key can only be used with distinct measures"
                )

            if values.get(
                "percentile"
            ) is not None and not measure_type.value.startswith("percentile"):
                raise ValueError("percentile can only be used with percentile measures")

            if values.get("precision") is not None and measure_type not in [
                LookerMeasureType.AVERAGE,
                LookerMeasureType.SUM,
            ]:
                raise ValueError(
                    "precision can only be used with average or sum measures"
                )

        return values

class LookerDimension(LookerViewElement):
    """Looker-specific data for a dimension on a model column

        looker:
            dimension:

    """
    primary_key: Optional[Union[bool, str]] = Field(default=None)
    convert_tz: Optional[Union[str, bool]] = Field(default=None)
    timeframes: Optional[List[str]] = Field(default=None)
    can_filter: Optional[Union[bool, str]] = Field(default=None)
    group_item_label: Optional[str] = Field(default=None)
    order_by_field: Optional[str] = Field(default=None)
    suggestable: Optional[bool] = Field(default=None)
    case_sensitive: Optional[bool] = Field(default=None)
    allow_fill: Optional[bool] = Field(default=None)
    required_access_grants: Optional[List[str]] = Field(default=None)
    html: Optional[str] = Field(default=None)
    sql: Optional[str] = None
    # fields: Optional[List['LookerDimension']] = None


    # @field_validator("timeframes", mode="before")
    # def check_enums(cls, values):
    #     if values is not None:
    #         if isinstance(values, list[str]):
    #             timeframes = values
    #             valid_timeframes = [
    #                 tf for tf in timeframes if isinstance(tf, LookerTimeFrame)
    #             ]
    #             if len(valid_timeframes) < len(timeframes):
    #                 invalid_timeframes = set(timeframes) - set(valid_timeframes)
    #                 warnings.warn(
    #                     f"Invalid timeframes: {invalid_timeframes}. "
    #                 )
    #                 values["timeframes"] = valid_timeframes

    #     return values


class ValidatedLookerDimension(LookerDimension):
    """Looker data for a dimension with validation."""

    @field_validator("convert_tz", "hidden", "primary_key", mode="after")
    @classmethod
    def bool_to_yesno(cls, value):
        if isinstance(value, bool):
            if value:
                return "yes"
            else:
                return "no"
        return value

class ValidatedLookerDimensionGroup(LookerDimension):
    """Looker data for a dimension group with validation."""

    @model_validator(mode="before")
    def convert_to_dimension_group(cls, values):
        if values.get("type") in (
            "date",
            "time"
        ):
            # Convert to dimension group
            values["type"] = "time"
            values["timeframes"] = LookerDateTimeframes.str_values()
            values["convert_tz"] = False

        elif values.get("type") in (
            "datetime",
            "timestamp",
        ):
            # Convert to dimension group
            values["type"] = "time"
            values["timeframes"] = LookerTimeTimeframes.str_values()
            values["convert_tz"] = True

        return values

    @field_validator("convert_tz", "hidden", mode="after")
    @classmethod
    def bool_to_yesno(cls, value):
        if isinstance(value, bool):
            if value:
                return "yes"
            else:
                return "no"
        return value

class LookerView(BaseModel):
    """Looker data for a view."""

    name: str
    label: Optional[str] = None
    sql_table_name: Optional[str] = None

    dimensions: Optional[List[ValidatedLookerDimension]] = Field(default=None)
    dimension_groups: Optional[List[ValidatedLookerDimensionGroup]] = Field(default=None)
    measures: Optional[List[LookerMeasure]] = Field(default=None)

    @model_validator(mode="before")
    def handle_time_and_dates(cls, values):
        """Convert time and date dimensions to dimension groups."""
        if values.get("dimensions"):
            dimensions = []
            dimensions_groups = []

            for dimension in values["dimensions"]:
                if dimension.get("type") in (
                    "date",
                    "datetime",
                    "timestamp",
                    "time",
                ):
                    dimensions_groups.append(dimension)
                else:
                    dimensions.append(dimension)
            values["dimensions"] = dimensions
            values["dimension_groups"] = dimensions_groups
        return values



class DotManipulation:
    """general . manipulation functions for adjusting strings to be used in looker"""

    def remove_dots(self, input_string: str) -> str:
        """replace all periods with a replacement string
        this is used to create unique names for joins
        """
        sign = "."
        replacement = "__"

        return input_string.replace(sign, replacement)

    def last_dot_only(self, input_string):
        """replace all but the last period with a replacement string
        IF there are multiple periods in the string
        this is used to create unique names for joins
        """
        sign = "."
        replacement = "__"

        # Splitting input_string into parts separated by sign (period)
        parts = input_string.split(sign)

        # If there's more than one part, we need to do replacements.
        if len(parts) > 1:
            # Joining all parts except for last with replacement,
            # and then adding back on final part.
            output_string = replacement.join(parts[:-1]) + sign + parts[-1]

            return output_string

        # If there are no signs at all or just one part,
        return input_string

    def textualize_dots(self, input_string: str) -> str:
        """Replace all periods with a human-readable " " """
        sign = "."
        replacement = " "

        return input_string.replace(sign, replacement)



from typing import List, Optional
from pydantic import BaseModel, model_validator
from looker_loader.models.looker import (
    LookerMeasure,
    LookerDimension,
)
from looker_loader.enums import LookerType

# Models for loading recipes, and for generating looker data from the combination of 
# the database and the recipes

class LookerRecipeDerivedDimension(LookerDimension):
    """A derived dimension in Looker"""
    suffix: str
    sql: Optional[str] = None
    html: Optional[str] = None
    measures: Optional[List[LookerMeasure]] = None
    variants: Optional[List['LookerRecipeDerivedDimension']] = None

class LookerRecipeDimension(LookerDimension):
    """A recipe dimension in Looker"""
    variants: Optional[List[LookerRecipeDerivedDimension]] = None
    measures: Optional[List[LookerMeasure]] = None
    fields: Optional[List[LookerDimension]] = None

class LookerMixtureMeasure(LookerMeasure):
    """A derived measure in Looker"""
    name: str
    sql: str
    html: Optional[str] = None

    @model_validator(mode="before")
    def fix_name(cls, values):
        values["name"] = f"m_{values.get('type')}_{values.get('parent_name')}"
        if values.get("sql") is None:
            values["sql"] = f"${{{values.get("parent_name")}}}"
        if values.get("group_label") is None:
            values["group_label"] = values.get("parent_group_label")
        if values.get("description") is None:
            values["description"] = f"{values.get('type')} of {values.get("parent_name")} : {values.get("parent_description")}"
        return values

class LookerMixtureDimension(LookerRecipeDimension):
    measures: Optional[List[LookerMixtureMeasure]] = None
    variants: Optional[List['LookerMixtureDimension']] = None
    fields: Optional[List['LookerMixtureDimension']] = None
    suffix: Optional[str] = None
    is_variant: Optional[bool] = None
    sql_table_name: Optional[str] = None

    @model_validator(mode="before")
    def create(cls, values):

        def apply(field : dict, list_params : List[str]):
            for param in list_params:
                if field.get(param) is None:
                    field[f"parent_{param}"] = values.get(param)
            return field

        if values.get("measures") is not None:
            inherited_children = []
            for child in values.get("measures", []):
                child = apply(child, ["name", "sql", "type", "group_label", "description", "tags"])
                inherited_children.append(child)
            values["measures"] = inherited_children

        if values.get("variants") is not None:
            inherited_children = []
            for child in values.get("variants", []):
                child = apply(child, ["name", "sql", "type", "group_label", "description", "tags"])
                inherited_children.append(child)
            values["variants"] = inherited_children
        return values

    @model_validator(mode="before")
    def fix_name(cls, values):
        if values.get("suffix") is not None:
            values["name"] = f"d_{values.get('parent_name')}_{values.get('suffix')}"
            if values.get("group_label") is None:
                values["group_label"] = values.get("parent_group_label")
            if values.get("description") is None:
                values["description"] = f"derived {values.get('suffix')} of {values.get("parent_name")} : {values.get("parent_description")}"
            values["type"] = values.get("parent_type", "string")
            values["sql"] = values.get("sql").replace("$x", f"${{{values.get('parent_name')}}}")
        return values

class LookerMixture(BaseModel):
    """A mixture of dimensions and measures in Looker"""
    name: str
    fields: Optional[List[LookerMixtureDimension]] = None
    measures: Optional[List[LookerMixtureMeasure]] = None
    dimensions: Optional[List[LookerMixtureDimension]] = None
    sql: Optional[str] = None
    sql_table_name: Optional[str] = None

class RecipeFilter(BaseModel):
    """a filter for a recipe"""

    type: Optional[LookerType] = None
    regex_include: Optional[str] = None
    regex_exclude: Optional[str] = None
    tags: Optional[List[str]] = None
    fields_include: Optional[List[str]] = None
    fields_exclude: Optional[List[str]] = None
    field_order: Optional[List[int]] = None

    @model_validator(mode="before")
    def check_at_least_one_field(cls, values):
        # Check if all values are None
        if all(value is None for value in values.values()):
            raise ValueError("At least one field must be specified.")
        return values

class Recipe(BaseModel):
    """
    A recipe for what to generate automatically in Looker
    - filters: filter what columns to apply the recipe to
    - dimension: metadata to add on the column
    - measures: measures to generate for the columns
    that are useful for that data
    """

    name: str
    filters: RecipeFilter
    dimension: Optional[LookerRecipeDimension] = None

class CookBook(BaseModel):
    """A cookbook of recipes"""

    # every recipe has a name
    recipes: List[Recipe]
